#!/usr/bin/bash

# vim:ft=sh:fdm=marker:fmr={,}

# archlabs installer library script file
# this file is not meant to be run directly
# sourcing this file in a non bash shell is not advised

########################################################
##               Partition functions                  ##
########################################################

format() {
    infobox "$_FSTitle" "\nFormatting: $1\n\nCommand: ${FS_CMDS[$2]}\n" 0
    ${FS_CMDS[$2]} $1 >/dev/null 2>$ERR
    check_for_errors "${FS_CMDS[$2]} $1"
}

decr_count() {
    # remove a partition from the dialog list and decrement the number partitions left
    local p="$1"
    PARTS="$(sed "s~${p} [0-9]*[G-M]~~; s~${p} [0-9]*\.[0-9]*[G-M]~~" <<< "$PARTS")"
    (( COUNT > 0 )) && (( COUNT-- ))
    return 0
}

enable_swap() {
    if [[ $1 == "$MNT/swapfile" ]]; then
        fallocate -l $SWAP_SIZE $1 2>$ERR
        check_for_errors "fallocate -l $SWAP_SIZE $1"
        chmod 600 $1 2>$ERR
        check_for_errors "chmod 600 $1"
    fi
    mkswap $1 >/dev/null 2>$ERR
    check_for_errors "mkswap $1"
    swapon $1 >/dev/null 2>$ERR
    check_for_errors "swapon $1"
    return 0
}

mount_partition() {
    local part="$1"
    local mount="${MNT}$2"
    local fs="$(lsblk -lno FSTYPE $part)"

    mkdir -p "$mount"

    if [[ ${FS_OPTS[$fs]} != "" && $part != "$BOOT_PART" ]] && select_mount_opts "$part" "$fs"; then
        mount -o $MNT_OPTS $part "$mount" 2>$ERR
        check_for_errors "mount -o $MNT_OPTS $part $mount"
    else
        mount $part "$mount" 2>$ERR
        check_for_errors "mount $part $mount"
    fi

    confirm_mount $part "$mount" || return 1
    check_cryptlvm "$part"
    return 0
}

confirm_mount() {
    local part="$1"
    local mount="$2"

    if [[ $mount == "$MNT" ]]; then
        local msg="Partition: $part\nMountpoint: / (root)"
    else
        local msg="Partition: $part\nMountpoint: ${mount#$MNT}"
    fi

    # partition failed to mount properly
    if ! [[ $(mount) =~ "$mount" ]]; then
        infobox "$_MntTitle" "$_MntFail\n$msg\n" 1
        return 1
    else
        # mount was successful
        infobox "$_MntTitle" "$_MntSucc\n$msg\n" 1
        decr_count "$part"
    fi

    return 0
}

find_partitions() {
    local str="$1"
    local err='NONE'

    # string of partitions as /TYPE/PART SIZE
    if [[ $IGNORE_DEV != "" ]]; then
        PARTS="$(lsblk -lno TYPE,NAME,SIZE |
            awk "/$str/"' && !'"/$IGNORE_DEV/"' {sub(/^part/, "/dev/");
                sub(/^lvm|^crypt/, "/dev/mapper/"); print $1$2 " " $3}')"
    else
        PARTS="$(lsblk -lno TYPE,NAME,SIZE |
            awk "/$str/"' {sub(/^part/, "/dev/") sub(/^lvm|^crypt/, "/dev/mapper/")
                print $1$2 " " $3}')"
    fi

    # number of partitions total
    COUNT=$(wc -l <<< "$PARTS")

    # ensure we have enough partitions for the system and action type
    case $str in
        'part|lvm|crypt') [[ $COUNT -eq 0 || ($SYS == 'UEFI' && $COUNT -lt 2) ]] && err="$_PartErrBody" ;;
        'part|crypt') (( COUNT == 0 )) && err="$_LvmPartErrBody" ;;
        'part|lvm') (( COUNT < 2 )) && err="$_LuksPartErrBody" ;;
    esac

    # if there aren't enough partitions show the error message
    if [[ $err != 'NONE' ]]; then
        msgbox "$_ErrTitle" "$err"
        return 1
    fi
    return 0
}

check_cryptlvm() {
    local part="$1"
    local devs="$(lsblk -lno NAME,FSTYPE,TYPE)"

    # Identify if $part is "crypt" (LUKS on LVM, or LUKS alone)
    if [[ $(lsblk -lno TYPE "$part") =~ 'crypt' ]]; then
        LUKS=1
        LUKS_NAME="${part#/dev/mapper/}"

        for dev in $(awk '/lvm/ && /crypto_LUKS/ {print "/dev/mapper/"$1}' <<< "$devs" | uniq); do
            if grep -q "$LUKS_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_DEV="$LUKS_DEV cryptdevice=$dev:$LUKS_NAME"
                LVM=1
                break
            fi
        done
        for dev in $(awk '/part/ && /crypto_LUKS/ {print "/dev/"$1}' <<< "$devs" | uniq); do
            if grep -q "$LUKS_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_UUID="$(lsblk -lno UUID,TYPE,FSTYPE "$dev" | awk '/part/ && /crypto_LUKS/ {print $1}')"
                LUKS_DEV="$LUKS_DEV cryptdevice=UUID=$LUKS_UUID:$LUKS_NAME"
                break
            fi
        done

    elif [[ $(lsblk -lno TYPE "$part") =~ 'lvm' ]]; then
        LVM=1
        VOLUME_NAME="${part#/dev/mapper/}"

        for dev in $(awk '/crypt/ && /lvm2_member/ {print "/dev/mapper/"$1}' <<< "$devs" | uniq); do
            if grep -q "$VOLUME_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_NAME="$(sed 's~/dev/mapper/~~g' <<< "$dev")"
                break
            fi
        done
        for dev in $(awk '/part/ && /crypto_LUKS/ {print "/dev/"$1}' <<< "$devs" | uniq); do
            if grep -q "$LUKS_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_UUID="$(lsblk -lno UUID,TYPE,FSTYPE "$dev" | awk '/part/ && /crypto_LUKS/ {print $1}')"
                LUKS_DEV="$LUKS_DEV cryptdevice=UUID=$LUKS_UUID:$LUKS_NAME"
                LUKS=1
                break
            fi
        done
    fi
}

auto_partition() {
    local device="$1"

    # confirm or bail
    yesno "$_PrepParts" "$_PartBody1 $device $_PartBody2" || return 0
    infobox "$_PrepParts" "\nAuto partitioning device: $device\n" 0

    swapoff -a  # make sure swap is disabled in case the device was used for swap

    local dev_info="$(parted -s $device print)"

    # walk the partitions on the device in reverse order and delete them
    for i in $(awk '/^ [1-9][0-9]?/ {print $1}' <<< "$dev_info" | sort -r); do
        parted -s $device rm $i >/dev/null 2>&1
    done

    # make sure we have the correct device table for the system type, gpt or msdos
    local newtable="gpt"
    local table="$(awk '/Table:/ {print $3}' <<< "$dev_info")"
    [[ $SYS == BIOS ]] && newtable="msdos"

    # if the current device table isn't correct run mklabel
    if [[ $table != "$newtable" ]]; then
        parted -s $device mklabel $newtable >/dev/null 2>&1
    fi

    # when device contains the string 'nvme' then add 'p' before the part number
    local nvme=""
    [[ $device =~ nvme ]] && nvme="p"

    local part_num=1
    BOOT_PART="$device${nvme}$part_num"  # set the boot partition label to the first partition

    if [[ $SYS == "BIOS" ]]; then
        parted -s $device mkpart primary ext4 1MiB 513MiB >/dev/null 2>&1
        mkfs.ext4 -q $BOOT_PART >/dev/null 2>&1
    else
        parted -s $device mkpart ESP fat32 1MiB 513MiB >/dev/null 2>&1
        mkfs.vfat -F32 $BOOT_PART >/dev/null 2>&1
    fi

    (( part_num++ ))                        # increment the partition number
    BOOT_DEVICE="$device"                   # only grub on BIOS systems uses this
    ROOT_PART="${device}${nvme}$part_num"   # set root partition label to the second partition

    parted -s $device mkpart primary ext4 513MiB 100% >/dev/null 2>&1
    mkfs.ext4 -q $ROOT_PART >/dev/null 2>&1

    tput civis
    sleep 0.5  # slow the process down, otherwise lsblk output can be wrong for some things

    echo -e "\nAuto partitioning complete.\n" > /tmp/.devlist
    lsblk $device -o NAME,MODEL,TYPE,FSTYPE,SIZE >> /tmp/.devlist
    dialog --cr-wrap --backtitle "$BT" --title " $_PrepParts " --textbox /tmp/.devlist 0 0
    return 0
}

setup_boot_device() {
    # set BOOT_DEVICE for syslinux on UEFI and grub on BIOS
    if [[ $BOOT_PART =~ nvme ]]; then
        BOOT_DEVICE="${BOOT_PART%p[1-9]}"
    else
        BOOT_DEVICE="${BOOT_PART%[1-9]}"
    fi

    BOOT_PART_NUM="${BOOT_PART: -1}"

    # setup the needed partition flags for boot on both system types
    parted -s $BOOT_DEVICE set $BOOT_PART_NUM boot on >/dev/null 2>&1
    if [[ $SYS == 'UEFI' ]]; then
        parted -s $BOOT_DEVICE set $BOOT_PART_NUM esp on >/dev/null 2>&1
    fi

    return 0
}

show_devices() {
    tput civis
    if [[ $IGNORE_DEV != "" ]]; then
        lsblk -o NAME,MODEL,TYPE,FSTYPE,SIZE,MOUNTPOINT |
            awk "!/$IGNORE_DEV/"' && /disk|part|lvm|crypt|NAME/ {print $0}' > /tmp/.devlist
    else
        lsblk -o NAME,MODEL,TYPE,FSTYPE,SIZE,MOUNTPOINT |
            awk '/disk|part|lvm|crypt|NAME/ {print $0}' > /tmp/.devlist
    fi
    dialog --cr-wrap --backtitle "$BT" --title " $_PrepShowDev " --textbox /tmp/.devlist 0 0
}

select_device() {
    local dev
    local msg
    if [[ $1 == 'boot' ]]; then
        msg="$_DevSelTitle for bootloader\n"
    else
        unmount_partitions
    fi

    if (( DEV_COUNT == 1 )) && [[ $SYS_DEVS ]]; then
        DEVICE="$(awk '{print $1}' <<< "$SYS_DEVS")"
        msg="\nOnly one device available$([[ $1 == 'boot' ]] && echo -n " for grub bootloader"):"
        infobox "$_DevSelTitle" "$msg $DEVICE\n" 1
    elif (( DEV_COUNT > 1 )); then
        tput civis
        DEVICE="$(dialog --cr-wrap --stdout --backtitle "$BT" --title " $_DevSelTitle " \
            --menu "${msg}$_DevSelBody" 0 0 0 $SYS_DEVS)"
        [[ $? != 0 || $DEVICE == "" ]] && return 1
    else
        msg="\nNo available devices for installation to use$([[ $1 == 'boot' ]] && echo -n " for grub bootloader")."
        msgbox "$_ErrTitle" "$msg\n$_Exit"
        die 1
    fi

    # if the device selected was for grub bootloader, set the BOOT_DEVICE
    # this is needed because grub uses the base device for BIOS, not the partition
    [[ $1 == 'boot' ]] && BOOT_DEVICE="$DEVICE"

    return 0
}

edit_partitions() {
    local device="$1"
    if [[ $device == "" ]]; then
        select_device 'root' || return 1
        device="$DEVICE"
    fi

    tput civis
    local choice
    choice="$(dialog --cr-wrap --stdout --backtitle "$BT" \
        --title " $_PartTitle " --menu "$_PartBody" 0 0 0 "$_PartAuto" "-" \
        $( ([[ $DISPLAY ]] && hash gparted >/dev/null 2>&1) && echo -n "gparted -") "cfdisk" "-"\
        "parted" "-" "$_PartWipe" "-")"
    [[ $? != 0 || $choice == "" ]] && return 1

    clear
    tput cnorm

    if [[ $choice != "$_PartWipe" && $choice != "$_PartAuto" ]]; then
        $choice $device
    elif [[ $choice == "$_PartWipe" ]]; then
        yesno "$_PartWipe" "$_PartBody1 $device $_PartWipeBody2" && { clear; wipe -Ifrev $device; }
        edit_partitions $device
    else
        # if auto_partition fails we need to re-initialize the variables, just to be sure
        auto_partition $device || { initialize_variables; return 1; }
    fi
}

########################################################
##                  Mount functions                   ##
########################################################

mount_main() {
    msgbox "$_PrepMount" "$_WarnMount"
    lvm_detect

    # prepare partition list PARTS for dialog
    unmount_partitions
    find_partitions 'part|lvm|crypt' || return 1

    # remove boot partition from dialog list if we auto partitioned one
    [[ $BOOT_PART != "" ]] && decr_count "$BOOT_PART"

    select_root_partition || return 1

    if [[ $BOOT_PART == "" ]]; then
        if [[ $SYS == "UEFI" ]]; then
            select_efi_partition || { BOOT_PART=""; return 1; }
        elif (( $COUNT > 0 )); then
            select_boot_partition || { BOOT_PART=""; return 1; }
        fi
    else
        infobox "$_PrepMount" "\nUsing boot partition: $BOOT_PART\n" 1
    fi

    select_boot_setup || { BOOTLDR=""; return 1; }

    if [[ $BOOT_PART != "" ]]; then
        setup_boot_device
        mount_partition "$BOOT_PART" "${BMNTS[$SYS-$BOOTLDR]}" || return 1
        SEPERATE_BOOT=true
    fi

    select_swap || return 1
    select_extra_partitions || return 1

    return 0
}

select_swap() {
    # Ask user to select partition or create swapfile
    tput civis
    SWAP="$(dialog --backtitle "$BT" --cr-wrap --stdout --title " $_SelSwpSetup " \
        --menu "$_SelSwpBody" 0 0 0 "$_SelSwpNone" "-" "$_SelSwpFile" "-" $PARTS)"
    [[ $? != 0 || $SWAP == "$_SelSwpNone" ]] && return 0

    if [[ $SWAP == "$_SelSwpFile" ]]; then
        tput cnorm
        SWAP_SIZE="$(getinput "$_SelSwpSetup" "$_SelSwpSize" "${SYS_MEM}M")"
        [[ $? != 0 || $SWAP_SIZE == "" ]] && return 1

        while ! [[ ${SWAP_SIZE:0:1} =~ [1-9] && ${SWAP_SIZE: -1} =~ (M|G) ]]; do
            msgbox "$_SelSwpSetup Error" "\n$_SelSwpErr $SWAP_SIZE\n"
            SWAP_SIZE="$(getinput "$_SelSwpSetup" "$_SelSwpSize" "${SYS_MEM}M")"
            [[ $? != 0 || $SWAP_SIZE == "" ]] && { break; return 1; }
        done

        enable_swap "$MNT/swapfile"
        SWAP="/swapfile"
    else
        enable_swap "$SWAP"
        decr_count "$SWAP"
    fi

    return 0
}

select_mount_opts() {
    local part="$1"
    local fs="$2"
    local title="${fs^} Mount Options"

    tput civis
    MNT_OPTS="$(dialog --cr-wrap --stdout --backtitle "$BT" --title " $title " \
        --checklist "$_MntBody" 0 0 0 ${FS_OPTS[$fs]} | sed 's/ /,/g; $s/,$//')"
    [[ $? != 0 || $MNT_OPTS == "" ]] && return 1

    if ! yesno "$title" "$_MntConfBody $MNT_OPTS\n"; then
        select_mount_opts "$part" "$fs" || return 1
    fi
    return 0
}

select_filesystem() {
    local part="$1"
    local cur_fs
    cur_fs="$(lsblk -lno FSTYPE $part)"
    tput civis

    local fs
    fs="$(dialog --cr-wrap --stdout --backtitle "$BT" --title " $_FSTitle: $part " \
        --menu "\nPartition: ${part}$([[ $cur_fs != "" ]] && echo -en "\nCurrent: ${cur_fs}")\n$_FSBody" 0 0 0 \
        $([[ $cur_fs != "" ]] && echo -n "$_Skip -") \
        "ext4" "${FS_CMDS[ext4]}"  "ext3" "${FS_CMDS[ext3]}" \
        "ext2" "${FS_CMDS[ext2]}"  "vfat" "${FS_CMDS[vfat]}" \
        "ntfs" "${FS_CMDS[ntfs]}"  "f2fs" "${FS_CMDS[f2fs]}" \
        "jfs" "${FS_CMDS[jfs]}"    "nilfs2" "${FS_CMDS[nilfs2]}" \
        "reiserfs" "${FS_CMDS[reiserfs]}" "xfs" "${FS_CMDS[xfs]}")"
    if [[ $fs == "$_Skip" ]]; then
        return 0
    elif [[ $fs == "" ]]; then
        return 1
    fi

    if yesno "$_FSTitle" "\nFormat $part as $fs?\n"; then
        format $part $fs
    else
        select_filesystem $part || return 1
    fi
    return 0
}

select_efi_partition() {
    tput civis
    if (( COUNT == 1 )); then
        BOOT_PART="$(awk 'NF > 0 {print $1}' <<< "$PARTS")"
        infobox "$_PrepMount" "$_OnlyOne for EFI: $BOOT_PART\n" 1
    else
        BOOT_PART="$(dialog --cr-wrap --stdout --backtitle "$BT" \
            --title " $_PrepMount " --menu "$_SelUefiBody" 0 0 0 $PARTS)"
        [[ $? != 0 || $BOOT_PART == "" ]] && return 1
    fi

    if grep -q 'fat' <<< "$(fsck -N "$BOOT_PART")"; then
        local msg="$_FormUefiBody $BOOT_PART $_FormUefiBody2"
        yesno "$_PrepMount" "$msg" "Format $BOOT_PART" "Do Not Format" "no" &&
            format "$BOOT_PART" "vfat"
    else
        format "$BOOT_PART" "vfat"
    fi
    return 0
}

select_boot_partition() {
    tput civis
    BOOT_PART="$(dialog --cr-wrap --stdout --backtitle "$BT" \
        --title " $_PrepMount " --menu "$_SelBiosBody" 0 0 0 "$_Skip" "-" $PARTS)"

    if [[ $BOOT_PART == "$_Skip" || $BOOT_PART == "" ]]; then
        BOOT_PART=""
    else
        if grep -q 'ext[34]' <<< "$(fsck -N "$BOOT_PART")"; then
            local msg="$_FormBiosBody $BOOT_PART $_FormUefiBody2"
            yesno "$_PrepMount" "$msg" "Format $BOOT_PART" "Skip Formatting" "no" &&
                format "$BOOT_PART" "ext4"
        else
            format "$BOOT_PART" "ext4"
        fi
    fi
    return 0
}

select_root_partition() {
    # if we used LUKS and no LVM or LUKS+LVM
    # remove the relevant partition labels from the list
    if (( LUKS == 1 && LVM == 0 )); then
        ROOT_PART="/dev/mapper/$LUKS_NAME"
        decr_count "$LUKS_PART"
    elif (( LUKS == 1 && LVM == 1 )); then
        decr_count "$LUKS_PART"
        for part in $(echo "$GROUP_PARTS"); do
            decr_count "$part"
        done
        ROOT_PART=""
    elif (( LUKS == 0 && LVM == 1 )); then
        for part in $(echo "$GROUP_PARTS"); do
            decr_count "$part"
        done
        ROOT_PART=""
    fi

    if [[ $COUNT -eq 1 && $ROOT_PART == "" ]]; then
        ROOT_PART="$(awk 'NF > 0 {print $1}' <<< "$PARTS")"
        infobox "$_PrepMount" "$_OnlyOne for root (/): $ROOT_PART\n" 1
    elif [[ $ROOT_PART == "" || $LVM -eq 1 ]]; then
        tput civis
        ROOT_PART="$(dialog --cr-wrap --stdout --backtitle "$BT" \
            --title " $_PrepMount " --menu "$_SelRootBody" 0 0 0 $PARTS)"
        [[ $? != 0 || $ROOT_PART == "" ]] && return 1
    else
        local msg="\nUsing $([[ $LUKS -eq 1 ]] && echo -n "encrypted ")root partition:"
        infobox "$_PrepMount" "$msg $ROOT_PART\n" 1
    fi

    select_filesystem "$ROOT_PART" || { ROOT_PART=""; return 1; }
    mount_partition "$ROOT_PART" || { ROOT_PART=""; return 1; }

    return 0
}

select_extra_partitions() {
    while (( COUNT > 0 )); do
        tput civis
        local part
        part="$(dialog --cr-wrap --stdout --backtitle "$BT" \
            --title " $_PrepMount " --menu "$_ExtPartBody" 0 0 0 "$_Done" "-" $PARTS)"
        [[ $? != 0 || $part == "$_Done" || $part == "" ]] && break

        # choose what filesystem and get mountpoint
        select_filesystem "$part" || { break; return 1; }
        select_mountpoint || { break; return 1; }

        # mount it
        mount_partition "$part" "$EXTRA_MNT" || { break; return 1; }

        EXTRA_MNTS="$EXTRA_MNTS $part: $EXTRA_MNT"

        # if the mountpoint was /usr add 'usr' to MKINIT_HOOKS
        [[ $EXTRA_MNT == "/usr" && $MKINIT_HOOKS != *usr* ]] && MKINIT_HOOKS="usr $MKINIT_HOOKS"
    done
    return 0
}

select_mountpoint() {
    tput cnorm
    EXTRA_MNT="$(getinput "$_PrepMount $part" "$_ExtPartBody1 /home /var\n" "/")"
    [[ $? != 0 || $EXTRA_MNT == "" ]] && return 1

    # bad mountpoint
    if [[ ${EXTRA_MNT:0:1} != "/" || ${#EXTRA_MNT} -le 1 || $EXTRA_MNT =~ \ |\' ]]; then
        msgbox "$_ErrTitle" "$_ExtErrBody"
        select_mountpoint || return 1
    fi

    return 0
}
